/*
  Name: pointlib.h
  Copyright: Copyright  Robert A Bradley 9/9/2005
  Author: Robert A Bradley
  Date: 09/09/2005 12:21
  Description: Point and Pointlist handling library
*/

#ifndef _POINTLIB_H
#define _POINTLIB_H
#include <stdlib.h>
#include <math.h>
#define POINTS_BLOCK 128
#define INUSE_BOOL int

typedef struct {
        double x;
        double y;
        double err_y;
        double fit_y;
        double weight_y;
        } Point;

typedef struct {
        /* Using an int for inUse is wasteful, but keeps us word aligned on 32
        bit hardware */
        INUSE_BOOL inUse;
        Point point;
        } PointlistItem;

typedef struct {
        int pointCount;
        void *next;
        void *prev;
        PointlistItem *points;
        } Pointlist;

/*
Pointlist* MakePointlist()

Returns a pointer to a Pointlist structure.  If there is insufficient memory,
the function returns NULL.
*/

Pointlist* MakePointlist() {
    Pointlist *points=(Pointlist*) malloc(sizeof(Pointlist));
    if (points==NULL) {
       return NULL;
    }
    points->pointCount=0;
    points->next=NULL;
    points->prev=NULL;
    points->points=(PointlistItem*) malloc(sizeof(PointlistItem)*POINTS_BLOCK);
    if (points->points==NULL) {
       free(points);
       return NULL;
    }
    int i;
    for (i=0;i<POINTS_BLOCK;i++) {
        points->points[i].inUse=0;
    }
    return points;
}

/*
void FreePointlist(Pointlist *pointsroot)

Frees a Pointlist structure and its children (pointed to by the "next" pointer).
*/

void FreePointlist(Pointlist *pointsroot) {
     if (pointsroot->prev!=NULL)
         pointsroot->prev=NULL; // Truncate list
     if (pointsroot->next!=NULL)
         FreePointlist((Pointlist*) pointsroot->next);
     free(pointsroot->points);
     free(pointsroot);
}

/*
void FreePointlist(Pointlist *pointsroot)

Frees a Pointlist segment, and relinks the list around it.
*/

void FreePointlistSegment(Pointlist *pl) {
     if (pl->prev!=NULL)
         ((Pointlist*) pl->prev)->next=(Pointlist*) pl->next;
     if (pl->next!=NULL)
         ((Pointlist*) pl->next)->prev=(Pointlist*) pl->prev;
     free(pl->points);
     free(pl);
}

/*
int AddPoint(Pointlist *points, Point p)

Adds a Point to a Pointlist structure.  If this fails, the function returns
a non-zero value.
*/

int AddPoint(Pointlist *points, Point p) {
       while (points->pointCount==POINTS_BLOCK) {
             if (points->next==NULL) {
                 points->next=MakePointlist();
                 ((Pointlist*) points->next)->prev=points;
                 if (points->next==NULL) {
                     return 1; // Out of memory
                 }
             }
             points=(Pointlist*) points->next;
       }
       int i;
       for (i=0;i<POINTS_BLOCK;i++) {
           if (!points->points[i].inUse)
               break;
       }
       PointlistItem *x=&(points->points[points->pointCount]);
       Point *d=&(x->point);
       d->x=p.x;
       d->y=p.y;
       d->err_y=p.err_y;
       //printf("%d, %f\n",points->pointCount,points->points[points->pointCount].x);
       x->inUse=1;
       points->pointCount++;
       return 0;
}

/*
void GetPointCount(Pointlist *pointsroot)

Counts the number of points in the whole linked list.
*/

long GetPointCount(Pointlist *pointsroot) {
    Pointlist *pl=pointsroot;
    long count=pl->pointCount;
    while (pl->next!=NULL) {
          pl=(Pointlist*) pl->next;
          count+=pl->pointCount;
    }
    return count;
}

/*
Point GetPoint(Pointlist *pointsroot, int index)

Returns the point at index index in the list (NULL if outside the list).
*/

Point GetPoint(Pointlist *pointsroot, int index) {
    int i=0;
    Pointlist *pl=pointsroot;
    while (pl->next!=NULL && index>(i+pl->pointCount)) {
          pl=(Pointlist*) pl->next;
          i+=pl->pointCount;
    }
    int x=index-i,ptr=0;
    if (pl->pointCount>=x && pl!=NULL) {
        while (x>=0) {
        //printf("GP: %d %d %d\n",x,ptr,pl->points[ptr].inUse);
              if (pl->points[ptr].inUse)
                  x--;
              ptr++;
        }
        ptr--;if (ptr<0) ptr=0;
        Point p=pl->points[ptr].point;
        //printf("Pointer: %d\n",ptr);
        return p;
    }
    Point p;
    return p;
}

/*
void RemovePoint(Pointlist *pointsroot, int index)

Removes (marks as free) the point at index.  If the block is empty, it is also freed
*/

void RemovePoint(Pointlist *pointsroot, int index) {
    int i=0;
    Pointlist *pl=pointsroot;
    Pointlist *plPrev=NULL;
    while (pl->next!=NULL && index>(i+pl->pointCount)) {
          plPrev=pl;
          pl=(Pointlist*) pl->next;
          i+=pl->pointCount;
    }
    int x=index-i,ptr=0;
    if (pl->pointCount>=x && pl!=NULL) {
        while (x>=0) {
              if (pl->points[ptr].inUse)
                  x--;
              ptr++;
        }
        ptr--;if (ptr<0) ptr=0;
        pl->points[ptr].inUse=0;
        pl->pointCount--;
        if (pl->pointCount<0)
            pl->pointCount=0;
        if (pl->pointCount==0 && pl->prev!=NULL) {
            FreePointlistSegment(pl);// Free if not root node
        }
    }
}

/*
Point* PointlistToArray(Pointlist *pointsroot)

Copies all the points into an array.  Returns NULL if malloc fails.
*/

Point* PointlistToArray(Pointlist *pointsroot) {
       int pCount=GetPointCount(pointsroot),x=0;
       int blocks=(int) ceil(pCount/(double) POINTS_BLOCK);
       int i,j;
       Pointlist *pl=pointsroot;
       PointlistItem *pli;
       Point *pts=(Point*) malloc(pCount*sizeof(Point));
       if (pts==NULL) {
          return NULL;
       }
       for (i=blocks;i>0;i--) {
           for (j=0;j<POINTS_BLOCK;j++) {
               pli=&(pl->points[j]);
               if (pli->inUse) {
                   pts[x]=pli->point;
                   x++;
               }
           }
           pl=(Pointlist*) pl->next;
       }
       return pts;
}

#ifdef _cplusplus
/* C++ functions - define _cplusplus to use */

/*
Point* PointListToArray(list<Point> *points)

Copies all the points in a C++ list of Points into an array.  Returns NULL if
malloc fails - free the array using free.
*/

#include <list>
using namespace std;
Point* PointListToArray(list<Point> *points) {
    int pCount=points->size();
    Point *pts=(Point*) malloc(pCount*sizeof(Point));
    if (pts==NULL) {
        return NULL;
    }
    list<Point>::iterator i=points->begin();
    int x=0;
    for (;i!=points->end();i++) {
        pts[x]=*i;
        x++;
    }
    return pts;
}
#endif
#endif
